import textdistance
import numpy as np
from nltk.translate.bleu_score import sentence_bleu
import math
import pandas as pd

def get_precision(y_true,y_pred):
    cnt=0
    for char in y_pred:
        if char in y_true:
            cnt+=1
    if len(y_pred)!=0:
        return cnt/len(y_pred)
    else:
        return 0
def get_recall(y_true,y_pred):
    cnt=0
    for char in y_pred:
        if char in y_true:
            cnt+=1
    if len(y_true)!=0:
        return cnt/len(y_true)
    else:
        return 0
def get_f1(precision,recall):
    if precision+recall!=0:
        return 2*(precision*recall)/(precision+recall)
    else:
        return 0
def get_precision_eng(y_true,y_pred):
    y_true=y_true.split(' ')
    y_pred=y_pred.split(' ')
    cnt=0
    for char in y_pred:
        if char in y_true:
            cnt+=1
    if len(y_pred)!=0:
        return cnt/len(y_pred)
    else:
        return 0
def get_recall_eng(y_true,y_pred):
    y_true=y_true.split(' ')
    y_pred=y_pred.split(' ')
    cnt=0
    for char in y_pred:
        if char in y_true:
            cnt+=1
    if len(y_true)!=0:
        return cnt/len(y_true)
    else:
        return 0
def get_f1_eng(precision,recall):
    if precision+recall!=0:
        return 2*(precision*recall)/(precision+recall)
    else:
        return 0
def get_bleu_score_eng(y_true,y_pred):
    reference=[y_true.split(' ')]
    candidate=y_pred.split(' ')
    gram_1=sentence_bleu(reference, candidate, weights=(1, 0, 0, 0))
    gram_2=sentence_bleu(reference, candidate, weights=(0, 1, 0, 0))
    gram_3=sentence_bleu(reference, candidate, weights=(0, 0, 1, 0))
    gram_4=sentence_bleu(reference, candidate, weights=(0, 0, 0, 1))
    res=0.0
    res_cnt=0
    n_gram_arr=[gram_1,gram_2,gram_3,gram_4]
    if len(candidate)>len(reference[0]):
        for x in n_gram_arr:
            if abs(x-0)>0.0001:
                res+=math.log(x)
                res_cnt+=1
        if res_cnt==0:
            return 0
        else:
            return math.exp(res/res_cnt)
    else:
        for x in n_gram_arr:
            if abs(x-0)>0.0001:
                res+=math.log(x)
                res_cnt+=1
        if res_cnt==0:
            return 0
        else:
            return math.exp(1-len(reference[0])/len(candidate))*math.exp(res/res_cnt)
def get_bleu_score(y_true,y_pred):
    reference=[list(y_true)]
    candidate=list(y_pred)
    gram_1=sentence_bleu(reference, candidate, weights=(1, 0, 0, 0))
    gram_2=sentence_bleu(reference, candidate, weights=(0, 1, 0, 0))
    gram_3=sentence_bleu(reference, candidate, weights=(0, 0, 1, 0))
    gram_4=sentence_bleu(reference, candidate, weights=(0, 0, 0, 1))
    res=0.0
    res_cnt=0
    n_gram_arr=[gram_1,gram_2,gram_3,gram_4]
    if len(candidate)>len(reference[0]):
        for x in n_gram_arr:
            if abs(x-0)>0.0001:
                res+=math.log(x)
                res_cnt+=1
        if res_cnt==0:
            return 0
        else:
            return math.exp(res/res_cnt)
    else:
        for x in n_gram_arr:
            if abs(x-0)>0.0001:
                res+=math.log(x)
                res_cnt+=1
        if res_cnt==0:
            return 0
        else:
            return math.exp(1-len(reference[0])/len(candidate))*math.exp(res/res_cnt)

def result_check_Val(dirstr,epochs,language):
    valPrdScore=[]
    for ep in range(epochs):
        with open(firDir+'/ValPredictionOnEpoch_'+str(ep+1)+'.txt','r',encoding='utf-8') as f:
            lines=f.readlines()
        # Levenshtein distance，normalized Levenshtein，Ratcliff Obershelp，Bleu，Precision，Recall，F1
        temscore=np.array([0.0,0.0,0.0,0.0,0.0,0.0,0.0])
        cnt=0
        for line in lines:
            if line[:4]=='true':
                cnt+=1
                temarr=line[:-1].split('\t')
                y_true=temarr[0].split(': ')[1]
                y_pred=temarr[1].split(': ')[1]
                if language=='chi':
                    ld=textdistance.levenshtein(y_pred,y_true)
                    nld=textdistance.levenshtein.normalized_similarity(y_pred,y_true)
                    ro=textdistance.ratcliff_obershelp(y_pred,y_true)
                    bleu=get_bleu_score(y_true,y_pred)
                    precision=get_precision(y_true,y_pred)
                    recall=get_recall(y_true,y_pred)
                    f1=get_f1(precision=precision,recall=recall)
                else:
                    ld=textdistance.levenshtein(y_pred,y_true)
                    nld=textdistance.levenshtein.normalized_similarity(y_pred,y_true)
                    ro=textdistance.ratcliff_obershelp(y_pred,y_true)
                    bleu=get_bleu_score_eng(y_true,y_pred)
                    precision=get_precision_eng(y_true,y_pred)
                    recall=get_recall_eng(y_true,y_pred)
                    f1=get_f1_eng(precision=precision,recall=recall)
                temscore[0]+=ld
                temscore[1]+=nld
                temscore[2]+=ro
                temscore[3]+=bleu
                temscore[4]+=precision
                temscore[5]+=recall
                temscore[6]+=f1
        temscore=temscore/cnt
        valPrdScore.append(temscore)
    ld_min=np.argmin([x[0] for x in valPrdScore])+1
    nld_min=np.argmax([x[1] for x in valPrdScore])+1
    ro_min=np.argmax([x[2] for x in valPrdScore])+1
    bleu_min=np.argmax([x[3] for x in valPrdScore])+1
    p_min=np.argmax([x[4] for x in valPrdScore])+1
    r_min=np.argmax([x[5] for x in valPrdScore])+1
    f1_min=np.argmax([x[6] for x in valPrdScore])+1
    best_ind=[ld_min,nld_min,ro_min,bleu_min,p_min,r_min,f1_min]
    return valPrdScore,best_ind

def result_check_Test(dirstr,language,best_ind):
    best_res=[]
    best_ind_cnt=0
    for epoch in best_ind:
        with open(firDir+'/TestPredictionOnEpoch_'+str(epoch)+'.txt','r',encoding='utf-8') as f:
            lines=f.readlines()
            # Levenshtein distance，normalized Levenshtein，Ratcliff Obershelp，Bleu，Precision，Recall，F1
            temscore=np.array([0.0,0.0,0.0,0.0,0.0,0.0,0.0])
            cnt=0
            for line in lines:
                if line[:4]=='true':
                    cnt+=1
                    temarr=line[:-1].split('\t')
                    y_true=temarr[0].split(': ')[1]
                    y_pred=temarr[1].split(': ')[1]
                    if language=='chi':
                        ld=textdistance.levenshtein(y_pred,y_true)
                        nld=textdistance.levenshtein.normalized_similarity(y_pred,y_true)
                        ro=textdistance.ratcliff_obershelp(y_pred,y_true)
                        bleu=get_bleu_score(y_true,y_pred)
                        precision=get_precision(y_true,y_pred)
                        recall=get_recall(y_true,y_pred)
                        f1=get_f1(precision=precision,recall=recall)
                    else:
                        ld=textdistance.levenshtein(y_pred,y_true)
                        nld=textdistance.levenshtein.normalized_similarity(y_pred,y_true)
                        ro=textdistance.ratcliff_obershelp(y_pred,y_true)
                        bleu=get_bleu_score_eng(y_true,y_pred)
                        precision=get_precision_eng(y_true,y_pred)
                        recall=get_recall_eng(y_true,y_pred)
                        f1=get_f1_eng(precision=precision,recall=recall)
                    temscore[0]+=ld
                    temscore[1]+=nld
                    temscore[2]+=ro
                    temscore[3]+=bleu
                    temscore[4]+=precision
                    temscore[5]+=recall
                    temscore[6]+=f1
            temscore=temscore/cnt
        best_res.append(temscore[best_ind_cnt])
        print(temscore)
        best_ind_cnt+=1
    return best_res

