#!/usr/bin/env python
# coding: utf-8

import re
import numpy as np
from bert4keras.backend import keras, K
from bert4keras.layers import Loss
from bert4keras.models import build_transformer_model
from bert4keras.tokenizers import Tokenizer, load_vocab
from bert4keras.optimizers import Adam
from bert4keras.snippets import sequence_padding
from bert4keras.snippets import DataGenerator, AutoRegressiveDecoder
from keras.models import Model
from sklearn.model_selection import KFold
from sklearn.metrics import accuracy_score
import pickle
import jieba
import textdistance
import random
from bert4keras.layers import *
from bert4keras.snippets import insert_arguments
from bert4keras.snippets import delete_arguments
from bert4keras.snippets import is_string
from keras.models import Model
import json
from nltk.translate.bleu_score import sentence_bleu
import pandas as pd
import argparse
import os
from EvaluateTools import *

# parameters

parser = argparse.ArgumentParser(description='setting parameters')
parser.add_argument("--mask-per", type=float, default=0.7)
parser.add_argument("--maxlen", type=int, default=256)
parser.add_argument("--batch-size", type=int, default=2)
parser.add_argument("--steps-per-epoch", type=int, default=400)
parser.add_argument("--epochs", type=int, default=70)
parser.add_argument("--full-name-maxlen", type=int, default=32)
args = parser.parse_args()

maxlen = args.maxlen
batch_size = args.batch_size
steps_per_epoch = args.steps_per_epoch
epochs = args.epochs
full_name_maxlen=args.full_name_maxlen
mask_per=args.mask_per

# result directory
firDir='../Results/MACF_RM_Journal'

def create_dir_not_exist(path):
    if not os.path.exists(path):
        os.mkdir(path)

create_dir_not_exist(firDir)

# tools
def M_token2ID(sentence,isFirstSen):
    ids=tokenizer.tokens_to_ids(tokenizer.tokenize(sentence))
    if isFirstSen:
        return ids
    else:
        return ids[1:]

# truncate context
def truncate_sequences(maxlen, index, sequences):
    while True:
        lengths = [len(s) for s in sequences]
        if sum(lengths) > maxlen:
            if len(sequences[index])>1:
                sequences[index].pop(-2)
            else:
                sequences.pop(index)
        else:
            return sequences

def M_encode(sentences,maxlen):
    ids,segs,tokens=[],[],[]
    tokens.append(tokenizer.tokenize(sentences[0]))
    tokens.extend([tokenizer.tokenize(sen)[1:] for sen in sentences[1:]])
    sentences=truncate_sequences(maxlen, -2, tokens)
    for i in range(len(sentences)):
        temids=tokenizer.tokens_to_ids(sentences[i])
        ids.extend(temids)
        segs.extend([i]*len(temids))
    return ids,segs

def M_encode_pred(sentences,maxlen):
    ids,segs,tokens=[],[],[]
    tokens.append(tokenizer.tokenize(sentences[0]))
    tokens.extend([tokenizer.tokenize(sen)[1:] for sen in sentences[1:]])
    sentences=truncate_sequences(maxlen, -1, tokens)
    for i in range(len(sentences)):
        temids=tokenizer.tokens_to_ids(sentences[i])
        ids.extend(temids)
        segs.extend([i]*len(temids))
    return ids,segs

def M_encode_ws(sentences,maxlen):
    ids,segs,tokens=[],[],[]
    tokens.append(tokenizer.tokenize(sentences[0]+'。')[:-1])
    tokens.extend([tokenizer.tokenize(sen+'。')[1:-1] for sen in sentences[1:-2]])
    tokens.append(tokenizer.tokenize(sentences[-2]+'。')[1:])
    tokens.append(tokenizer.tokenize(sentences[-1])[1:])
    sentences=truncate_sequences(maxlen, -2, tokens)
    for i in range(len(sentences)):
        temids=tokenizer.tokens_to_ids(sentences[i])
        ids.extend(temids)
        segs.extend([i]*len(temids))
    return ids,segs

def M_encode_ws_pred(sentences,maxlen):
    ids,segs,tokens=[],[],[]
    tokens.append(tokenizer.tokenize(sentences[0]+'。')[:-1])
    tokens.extend([tokenizer.tokenize(sen+'。')[1:-1] for sen in sentences[1:-2]])
    tokens.append(tokenizer.tokenize(sentences[-1])[1:])
    sentences=truncate_sequences(maxlen, -1, tokens)
    for i in range(len(sentences)):
        temids=tokenizer.tokens_to_ids(sentences[i])
        ids.extend(temids)
        segs.extend([i]*len(temids))
    return ids,segs

def get_mask_matrix(matr):
    tem=[]
    for arr in matr:
        tem.append(np.floor_divide(arr,np.max(arr)))
    return np.array(tem)

def data2sentences(data,lang):
    res=[data[2]]
    if lang=='eng':
        temarr=data[0].split('.')
    elif lang=='chi':
        temarr=data[0].split('。')
    for sen in temarr:
        if sen!='':
            res.append(sen)
    res.append(data[1])
    return res

def data2sentences_DG(data,lang):
    res=[data[2]]
    if lang=='eng':
        temarr=data[0].split('.')
    elif lang=='chi':
        temarr=data[0].split('。')
    for sen in temarr:
        if sen!='':
            res.append(sen)
    return res

def get_custom_segment(segs):
    maxval=max(segs)
    allLen=len(segs)
    segs=np.array(segs)
    abblen=len(segs[segs==0])
    fnlen=len(segs[segs==maxval])
    res=abblen*[1]+(allLen-abblen-fnlen)*[2]+fnlen*[3]
    return list(res)

def get_custom_segment_pred(segs):
    allLen=len(segs)
    segs=np.array(segs)
    abblen=len(segs[segs==0])
    res=abblen*[1]+(allLen-abblen)*[2]
    return res

def get_random_num(start,end,num):
    res=[]
    while len(res)<num:
        temnum=random.randint(start,end)
        if temnum not in res:
            res.append(temnum)
    return res
def get_mask_arr(length,maxvalNum,per):
    res=[True]*length
    poslis=get_random_num(0,maxvalNum-1,np.ceil(maxvalNum*per))
    for pos in poslis:
        res[length-maxvalNum+pos]=False
    return res
def generate_RanM(segs,per):
    res=[]
    for seg in segs:
        tem=[]
        seg=np.array(seg)
        maxvalNum=len(seg[seg==3])
        zeroNum=len(seg[seg==0])
        ind=0
        maksArr=get_mask_arr(len(seg)-zeroNum,maxvalNum,per)
        for val in seg:
            if val==1:
                tem.append(list(seg[:len(seg)-zeroNum]<3)+[False]*zeroNum)
            elif val==2:
                tem.append(maksArr+[False]*zeroNum)
            elif val==3:
                ind+=1
                tem.append([True]*(len(seg)-maxvalNum-zeroNum)+ind*[True]+(maxvalNum-ind+zeroNum)*[False])
            elif val==0:
                tem.append([True]*len(seg))
        res.append(tem)
    return np.array(res)

# reading data
Journal_Data=pd.read_csv('../Data/JOURNAL.csv',encoding='utf-8')
JFN=list(Journal_Data['Full-name'])
JA=list(Journal_Data['Abbreviation'])
JC=list(Journal_Data['Context'])
data_list=[]
for i in range(len(JA)):
    data_list.append([JC[i].replace(JFN[i],JA[i]),JFN[i],JA[i]])
data_list=np.array(data_list)

# data_list:[[context,full-name,abbreviation]*]
data_list=np.array(data_list)

# BERT-wwm path
config_path = '../BERT/wwm_cased_L-24_H-1024_A-16/bert_config.json'
checkpoint_path = '../BERT/wwm_cased_L-24_H-1024_A-16/bert_model.ckpt'
dict_path = '../BERT/wwm_cased_L-24_H-1024_A-16/vocab.txt'

# load vocabulary
token_dict, keep_tokens = load_vocab(
    dict_path=dict_path,
    simplified=True,
    startswith=['[PAD]', '[UNK]', '[CLS]', '[SEP]'],
)
# tokenizer for word segmentation
tokenizer = Tokenizer(token_dict)

# generate data for training
class data_generator(DataGenerator):
    def __iter__(self, random=False):
        batch_token_ids, batch_segment_ids, batch_custom_segments = [], [], []
        for is_end, data in self.sample(random):
            sentences=data2sentences(data,'eng')
            token_ids, segment_ids = M_encode(sentences,maxlen)
            custom_segment=get_custom_segment(segment_ids)
            segment_ids=list(np.floor_divide(segment_ids,np.max(segment_ids)))
            batch_token_ids.append(token_ids)
            batch_segment_ids.append(segment_ids)
            batch_custom_segments.append(custom_segment)
            if len(batch_token_ids) == self.batch_size or is_end:
                batch_token_ids = sequence_padding(batch_token_ids,length=maxlen)
                batch_segment_ids = sequence_padding(batch_segment_ids,length=maxlen)
                batch_custom_segments = sequence_padding(batch_custom_segments,length=maxlen)
                batch_mask = generate_RanM(batch_custom_segments,mask_per)
                yield [batch_token_ids, batch_segment_ids, batch_mask], None
                batch_token_ids, batch_segment_ids,batch_mask,batch_custom_segments = [], [], [], []

# Loss
class CrossEntropy(Loss):
    def compute_loss(self, inputs, mask=None):
        y_true, y_mask, _, y_pred = inputs
        y_true = y_true[:, 1:]
        y_mask = y_mask[:, 1:]
        y_pred = y_pred[:, :-1]
        print(y_pred)
        loss = K.sparse_categorical_crossentropy(y_true, y_pred)
        loss = K.sum(loss * y_mask) / K.sum(y_mask)
        return loss

def generate_AccM(segs):
    res=[]
    for seg in segs:
        tem=[]
        seg=np.array(seg)
        maxvalNum=len(seg[seg==3])
        zeroNum=len(seg[seg==0])
        ind=0
        for val in seg:
            if val==1:
                tem.append(list(seg[:len(seg)-zeroNum]<3)+[False]*zeroNum)
            elif val==2:
                tem.append([True]*(len(seg)-zeroNum)+[False]*zeroNum)
            elif val==3:
                ind+=1
                tem.append([True]*(len(seg)-maxvalNum-zeroNum)+ind*[True]+(maxvalNum-ind+zeroNum)*[False])
            elif val==0:
                tem.append([True]*len(seg))
        res.append(tem)
    return np.array(res)

# custom mask
class Custom_UniLM_Mask(object):
    def compute_attention_bias(self, inputs=None):
        if self.attention_bias is None:
            def unilm_mask(s):
                mask = s
                mask = K.cast(mask, K.floatx())
                return -(1 - mask[:, None]) * 1e12

            self.attention_bias = self.apply(
                inputs=self.inputs[2],
                layer=Lambda,
                function=unilm_mask,
                name='Attention-UniLM-Mask'
            )

        return self.attention_bias

def extend_with_custom_unified_language_model(BaseModel):
    class Custom_UnifiedLanguageModel(Custom_UniLM_Mask, BaseModel):
        def __init__(self, *args, **kwargs):
            super(Custom_UnifiedLanguageModel, self).__init__(*args, **kwargs)
            self.with_mlm = self.with_mlm or True

    return Custom_UnifiedLanguageModel


# Transformer
class Transformer(object):
    def __init__(
        self,
        vocab_size,
        hidden_size,
        num_hidden_layers,
        num_attention_heads,
        intermediate_size,
        hidden_act,
        dropout_rate=None,
        embedding_size=None,
        attention_head_size=None,
        attention_key_size=None,
        sequence_length=None,
        keep_tokens=None,
        compound_tokens=None,
        residual_attention_scores=False,
        ignore_invalid_weights=False,
        layers=None,
        prefix=None,
        name=None,
        **kwargs
    ):
        if keep_tokens is not None:
            vocab_size = len(keep_tokens)
        if compound_tokens is not None:
            vocab_size += len(compound_tokens)
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.num_hidden_layers = num_hidden_layers
        self.num_attention_heads = num_attention_heads
        self.attention_head_size = attention_head_size or hidden_size // num_attention_heads
        self.attention_key_size = attention_key_size or self.attention_head_size
        self.intermediate_size = intermediate_size
        self.dropout_rate = dropout_rate or 0
        self.hidden_act = hidden_act
        self.embedding_size = embedding_size or hidden_size
        self.sequence_length = sequence_length
        self.keep_tokens = keep_tokens
        self.compound_tokens = compound_tokens
        self.attention_bias = None
        self.position_bias = None
        self.attention_scores = None
        self.residual_attention_scores = residual_attention_scores
        self.ignore_invalid_weights = ignore_invalid_weights
        self.layers = {} if layers is None else layers
        self.prefix = prefix or ''
        self.name = name
        self.built = False

    def build(
        self,
        attention_caches=None,
        layer_norm_cond=None,
        layer_norm_cond_hidden_size=None,
        layer_norm_cond_hidden_act=None,
        additional_input_layers=None,
        **kwargs
    ):
        if self.built:
            return None
        inputs = self.get_inputs()
        self.set_inputs(inputs, additional_input_layers)
        self.attention_caches = attention_caches or {}
        self.layer_norm_conds = [
            layer_norm_cond,
            layer_norm_cond_hidden_size,
            layer_norm_cond_hidden_act or 'linear',
        ]
        outputs = self.call(inputs)
        self.set_outputs(outputs)
        self.model = Model(self.inputs, self.outputs, name=self.name)
        self.built = True

    def call(self, inputs):
        outputs = self.apply_embeddings(inputs)
        for i in range(self.num_hidden_layers):
            outputs = self.apply_main_layers(outputs, i)
        outputs = self.apply_final_layers(outputs)
        return outputs

    def prefixed(self, name):
        if name is not None:
            return self.prefix + name

    def apply(self, inputs=None, layer=None, arguments=None, **kwargs):
        if layer is Dropout and self.dropout_rate == 0:
            return inputs

        if layer is MultiHeadAttention and self.residual_attention_scores:
            kwargs['return_attention_scores'] = True

        arguments = arguments or {}
        name = self.prefixed(kwargs.get('name'))
        kwargs['name'] = name
        if name not in self.layers:
            layer = layer(**kwargs)
            name = layer.name
            self.layers[name] = layer

        if inputs is None:
            return self.layers[name]
        else:
            if isinstance(self.layers[name], MultiHeadAttention):
                if name in self.attention_caches:
                    k_cache, v_cache = self.attention_caches[name]
                    k_name, v_name = name + '-Cached-Key', name + '-Cached-Value'
                    k = Concatenate1D(name=k_name)([k_cache, inputs[1]])
                    v = Concatenate1D(name=v_name)([v_cache, inputs[2]])
                    inputs = inputs[:1] + [k, v] + inputs[3:]
                if self.residual_attention_scores:
                    if self.attention_scores is not None:
                        if arguments.get('a_bias'):
                            a_bias = Add(name=name + '-Attention-Bias'
                                        )([inputs[3], self.attention_scores])
                        else:
                            a_bias = self.attention_scores
                        inputs = inputs[:3] + [a_bias] + inputs[4:]
                        arguments['a_bias'] = True
                    o, a = self.layers[name](inputs, **arguments)
                    self.attention_scores = a
                    return o
            return self.layers[name](inputs, **arguments)

    def get_inputs(self):
        raise NotImplementedError

    def apply_embeddings(self, inputs):
        raise NotImplementedError

    def apply_main_layers(self, inputs, index):
        raise NotImplementedError

    def apply_final_layers(self, inputs):
        raise NotImplementedError

    def compute_attention_bias(self, inputs=None):
        return self.attention_bias

    def compute_position_bias(self, inputs=None):
        return self.position_bias

    def set_inputs(self, inputs, additional_input_layers=None):
        if inputs is None:
            inputs = []
        elif not isinstance(inputs, list):
            inputs = [inputs]

        inputs = inputs[:]
        if additional_input_layers is not None:
            if not isinstance(additional_input_layers, list):
                additional_input_layers = [additional_input_layers]
            inputs.extend(additional_input_layers)

        self.inputs = inputs
        if len(inputs) > 1:
            self.input = inputs
        else:
            self.input = inputs[0]

    def set_outputs(self, outputs):
        if not isinstance(outputs, list):
            outputs = [outputs]

        outputs = outputs[:]
        self.outputs = outputs
        if len(outputs) > 1:
            self.output = outputs
        else:
            self.output = outputs[0]

    @property
    def initializer(self):
        return keras.initializers.TruncatedNormal(stddev=0.02)

    def simplify(self, inputs):
        inputs = [i for i in inputs if i is not None]
        if len(inputs) == 1:
            inputs = inputs[0]

        return inputs

    def load_embeddings(self, embeddings):
        embeddings = embeddings.astype(K.floatx())

        if self.keep_tokens is not None:
            embeddings = embeddings[self.keep_tokens]

        if self.compound_tokens is not None:
            ext_embeddings = []
            for item in self.compound_tokens:
                if isinstance(item, list):
                    item = (item, [1] * len(item))
                ext_embeddings.append(
                    np.average(embeddings[item[0]], 0, item[1])
                )
            embeddings = np.concatenate([embeddings, ext_embeddings], 0)

        return embeddings

    def load_variable(self, checkpoint, name):
        if isinstance(checkpoint, dict):
            return checkpoint[name]
        else:
            return tf.train.load_variable(checkpoint, name)

    def create_variable(self, name, value, dtype=None):
        dtype = dtype or K.floatx()
        return K.variable(
            self.initializer(value.shape, dtype), dtype, name=name
        ), value

    def variable_mapping(self):
        return {}

    def load_weights_from_checkpoint(self, checkpoint, mapping=None):
        mapping = mapping or self.variable_mapping()
        mapping = {self.prefixed(k): v for k, v in mapping.items()}
        mapping = {k: v for k, v in mapping.items() if k in self.layers}

        weight_value_pairs = []
        for layer, variables in mapping.items():
            layer = self.layers[layer]
            weights = layer.trainable_weights
            values = []

            for v in variables:
                try:
                    values.append(self.load_variable(checkpoint, v))
                except Exception as e:
                    if self.ignore_invalid_weights:
                        values.append(None)
                        print('%s, but ignored.' % e.message)
                    else:
                        raise e

            if isinstance(layer, MultiHeadAttention):
                count = 2
                if layer.use_bias:
                    count += 2
                heads = self.num_attention_heads
                head_size = self.attention_head_size
                key_size = self.attention_key_size
                W = np.linalg.qr(np.random.randn(key_size, head_size))[0].T
                if layer.attention_scale:
                    W = W * key_size**0.25 / head_size**0.25
                for i in range(count):
                    w, v = weights[i], values[i]
                    if v is None:
                        continue
                    w_shape, v_shape = K.int_shape(w), v.shape
                    if w_shape[-1] != v_shape[-1]:
                        pre_shape = w_shape[:-1]
                        v = v.reshape(pre_shape + (heads, head_size))
                        v = np.dot(v, W)
                        v = v.reshape(pre_shape + (heads * key_size,))
                        values[i] = v

            weight_value_pairs.extend([
                (w, v) for w, v in zip(weights, values) if v is not None
            ])

        K.batch_set_value(weight_value_pairs)

    def save_weights_as_checkpoint(self, filename, mapping=None, dtype=None):
        mapping = mapping or self.variable_mapping()
        mapping = {self.prefixed(k): v for k, v in mapping.items()}
        mapping = {k: v for k, v in mapping.items() if k in self.layers}

        with tf.Graph().as_default():
            all_variables, all_values = [], []
            for layer, variables in mapping.items():
                layer = self.layers[layer]
                values = K.batch_get_value(layer.trainable_weights)
                for name, value in zip(variables, values):
                    variable, value = self.create_variable(name, value, dtype)
                    all_variables.append(variable)
                    all_values.append(value)
            with tf.Session() as sess:
                K.batch_set_value(zip(all_variables, all_values))
                saver = tf.train.Saver()
                saver.save(sess, filename)

# BERT
class BERT(Transformer):
    def __init__(
        self,
        max_position,
        segment_vocab_size=2,
        with_pool=False,
        with_nsp=False,
        with_mlm=False,
        hierarchical_position=None,
        custom_position_ids=False,
        shared_segment_embeddings=False,
        **kwargs
    ):
        super(BERT, self).__init__(**kwargs)
        self.max_position = max_position
        self.segment_vocab_size = segment_vocab_size
        self.with_pool = with_pool
        self.with_nsp = with_nsp
        self.with_mlm = with_mlm
        self.hierarchical_position = hierarchical_position
        self.custom_position_ids = custom_position_ids
        self.shared_segment_embeddings = shared_segment_embeddings
        if self.with_nsp and not self.with_pool:
            self.with_pool = True

    def get_inputs(self):
        x_in = self.apply(
            layer=Input, shape=(self.sequence_length,), name='Input-Token'
        )
        inputs = [x_in]

        if self.segment_vocab_size > 0:
            s_in = self.apply(
                layer=Input,
                shape=(self.sequence_length,),
                name='Input-Segment'
            )
            inputs.append(s_in)

        if self.segment_vocab_size > 0:
            m_in = self.apply(
                layer=Input,
                shape=(self.sequence_length,None,),
                name='Input-Custom-Mask'
            )
            inputs.append(m_in)

        if self.custom_position_ids:
            p_in = self.apply(
                layer=Input,
                shape=(self.sequence_length,),
                name='Input-Position'
            )
            inputs.append(p_in)

        return inputs

    def apply_embeddings(self, inputs):
        inputs = inputs[:]
        x = inputs.pop(0)
        if self.segment_vocab_size > 0:
            s = inputs.pop(0)
        if self.custom_position_ids:
            p = inputs.pop(0)
        else:
            p = None
        z = self.layer_norm_conds[0]

        x = self.apply(
            inputs=x,
            layer=Embedding,
            input_dim=self.vocab_size,
            output_dim=self.embedding_size,
            embeddings_initializer=self.initializer,
            mask_zero=True,
            name='Embedding-Token'
        )
        if self.segment_vocab_size > 0:
            if self.shared_segment_embeddings:
                name = 'Embedding-Token'
            else:
                name = 'Embedding-Segment'
            s = self.apply(
                inputs=s,
                layer=Embedding,
                input_dim=self.segment_vocab_size,
                output_dim=self.embedding_size,
                embeddings_initializer=self.initializer,
                name=name
            )
            x = self.apply(
                inputs=[x, s], layer=Add, name='Embedding-Token-Segment'
            )
        x = self.apply(
            inputs=self.simplify([x, p]),
            layer=PositionEmbedding,
            input_dim=self.max_position,
            output_dim=self.embedding_size,
            merge_mode='add',
            hierarchical=self.hierarchical_position,
            embeddings_initializer=self.initializer,
            custom_position_ids=self.custom_position_ids,
            name='Embedding-Position'
        )
        x = self.apply(
            inputs=self.simplify([x, z]),
            layer=LayerNormalization,
            conditional=(z is not None),
            hidden_units=self.layer_norm_conds[1],
            hidden_activation=self.layer_norm_conds[2],
            hidden_initializer=self.initializer,
            name='Embedding-Norm'
        )
        x = self.apply(
            inputs=x,
            layer=Dropout,
            rate=self.dropout_rate,
            name='Embedding-Dropout'
        )
        if self.embedding_size != self.hidden_size:
            x = self.apply(
                inputs=x,
                layer=Dense,
                units=self.hidden_size,
                kernel_initializer=self.initializer,
                name='Embedding-Mapping'
            )

        return x

    def apply_main_layers(self, inputs, index):
        x = inputs
        z = self.layer_norm_conds[0]

        attention_name = 'Transformer-%d-MultiHeadSelfAttention' % index
        feed_forward_name = 'Transformer-%d-FeedForward' % index
        attention_mask = self.compute_attention_bias(index)

        xi, x, arguments = x, [x, x, x], {'a_bias': None}
        if attention_mask is not None:
            arguments['a_bias'] = True
            x.append(attention_mask)

        x = self.apply(
            inputs=x,
            layer=MultiHeadAttention,
            arguments=arguments,
            heads=self.num_attention_heads,
            head_size=self.attention_head_size,
            out_dim=self.hidden_size,
            key_size=self.attention_key_size,
            kernel_initializer=self.initializer,
            name=attention_name
        )
        x = self.apply(
            inputs=x,
            layer=Dropout,
            rate=self.dropout_rate,
            name='%s-Dropout' % attention_name
        )
        x = self.apply(
            inputs=[xi, x], layer=Add, name='%s-Add' % attention_name
        )
        x = self.apply(
            inputs=self.simplify([x, z]),
            layer=LayerNormalization,
            conditional=(z is not None),
            hidden_units=self.layer_norm_conds[1],
            hidden_activation=self.layer_norm_conds[2],
            hidden_initializer=self.initializer,
            name='%s-Norm' % attention_name
        )

        xi = x
        x = self.apply(
            inputs=x,
            layer=FeedForward,
            units=self.intermediate_size,
            activation=self.hidden_act,
            kernel_initializer=self.initializer,
            name=feed_forward_name
        )
        x = self.apply(
            inputs=x,
            layer=Dropout,
            rate=self.dropout_rate,
            name='%s-Dropout' % feed_forward_name
        )
        x = self.apply(
            inputs=[xi, x], layer=Add, name='%s-Add' % feed_forward_name
        )
        x = self.apply(
            inputs=self.simplify([x, z]),
            layer=LayerNormalization,
            conditional=(z is not None),
            hidden_units=self.layer_norm_conds[1],
            hidden_activation=self.layer_norm_conds[2],
            hidden_initializer=self.initializer,
            name='%s-Norm' % feed_forward_name
        )

        return x

    def apply_final_layers(self, inputs):
        x = inputs
        z = self.layer_norm_conds[0]
        outputs = [x]

        if self.with_pool:
            x = outputs[0]
            x = self.apply(
                inputs=x,
                layer=Lambda,
                function=lambda x: x[:, 0],
                name='Pooler'
            )
            pool_activation = 'tanh' if self.with_pool is True else self.with_pool
            x = self.apply(
                inputs=x,
                layer=Dense,
                units=self.hidden_size,
                activation=pool_activation,
                kernel_initializer=self.initializer,
                name='Pooler-Dense'
            )
            if self.with_nsp:
                x = self.apply(
                    inputs=x,
                    layer=Dense,
                    units=2,
                    activation='softmax',
                    kernel_initializer=self.initializer,
                    name='NSP-Proba'
                )
            outputs.append(x)

        if self.with_mlm:
            x = outputs[0]
            x = self.apply(
                inputs=x,
                layer=Dense,
                units=self.embedding_size,
                activation=self.hidden_act,
                kernel_initializer=self.initializer,
                name='MLM-Dense'
            )
            x = self.apply(
                inputs=self.simplify([x, z]),
                layer=LayerNormalization,
                conditional=(z is not None),
                hidden_units=self.layer_norm_conds[1],
                hidden_activation=self.layer_norm_conds[2],
                hidden_initializer=self.initializer,
                name='MLM-Norm'
            )
            x = self.apply(
                inputs=x,
                layer=Embedding,
                arguments={'mode': 'dense'},
                name='Embedding-Token'
            )
            x = self.apply(inputs=x, layer=BiasAdd, name='MLM-Bias')
            mlm_activation = 'softmax' if self.with_mlm is True else self.with_mlm
            x = self.apply(
                inputs=x,
                layer=Activation,
                activation=mlm_activation,
                name='MLM-Activation'
            )
            outputs.append(x)

        if len(outputs) == 1:
            outputs = outputs[0]
        elif len(outputs) == 2:
            outputs = outputs[1]
        else:
            outputs = outputs[1:]

        return outputs

    def load_variable(self, checkpoint, name):
        variable = super(BERT, self).load_variable(checkpoint, name)
        if name in [
            'bert/embeddings/word_embeddings',
            'cls/predictions/output_bias',
        ]:
            return self.load_embeddings(variable)
        elif name == 'cls/seq_relationship/output_weights':
            return variable.T
        else:
            return variable

    def create_variable(self, name, value, dtype=None):
        if name == 'cls/seq_relationship/output_weights':
            value = value.T
        return super(BERT, self).create_variable(name, value, dtype)

    def variable_mapping(self):
        mapping = {
            'Embedding-Token': ['bert/embeddings/word_embeddings'],
            'Embedding-Segment': ['bert/embeddings/token_type_embeddings'],
            'Embedding-Position': ['bert/embeddings/position_embeddings'],
            'Embedding-Norm': [
                'bert/embeddings/LayerNorm/beta',
                'bert/embeddings/LayerNorm/gamma',
            ],
            'Embedding-Mapping': [
                'bert/encoder/embedding_hidden_mapping_in/kernel',
                'bert/encoder/embedding_hidden_mapping_in/bias',
            ],
            'Pooler-Dense': [
                'bert/pooler/dense/kernel',
                'bert/pooler/dense/bias',
            ],
            'NSP-Proba': [
                'cls/seq_relationship/output_weights',
                'cls/seq_relationship/output_bias',
            ],
            'MLM-Dense': [
                'cls/predictions/transform/dense/kernel',
                'cls/predictions/transform/dense/bias',
            ],
            'MLM-Norm': [
                'cls/predictions/transform/LayerNorm/beta',
                'cls/predictions/transform/LayerNorm/gamma',
            ],
            'MLM-Bias': ['cls/predictions/output_bias'],
        }

        for i in range(self.num_hidden_layers):
            prefix = 'bert/encoder/layer_%d/' % i
            mapping.update({
                'Transformer-%d-MultiHeadSelfAttention' % i: [
                    prefix + 'attention/self/query/kernel',
                    prefix + 'attention/self/query/bias',
                    prefix + 'attention/self/key/kernel',
                    prefix + 'attention/self/key/bias',
                    prefix + 'attention/self/value/kernel',
                    prefix + 'attention/self/value/bias',
                    prefix + 'attention/output/dense/kernel',
                    prefix + 'attention/output/dense/bias',
                ],
                'Transformer-%d-MultiHeadSelfAttention-Norm' % i: [
                    prefix + 'attention/output/LayerNorm/beta',
                    prefix + 'attention/output/LayerNorm/gamma',
                ],
                'Transformer-%d-FeedForward' % i: [
                    prefix + 'intermediate/dense/kernel',
                    prefix + 'intermediate/dense/bias',
                    prefix + 'output/dense/kernel',
                    prefix + 'output/dense/bias',
                ],
                'Transformer-%d-FeedForward-Norm' % i: [
                    prefix + 'output/LayerNorm/beta',
                    prefix + 'output/LayerNorm/gamma',
                ],
            })

        return mapping

def build_transformer_model(
    config_path=None,
    checkpoint_path=None,
    model='bert',
    application='encoder',
    return_keras_model=True,
    **kwargs
):
    configs = {}
    if config_path is not None:
        configs.update(json.load(open(config_path)))
    configs.update(kwargs)
    if 'max_position' not in configs:
        configs['max_position'] = configs.get('max_position_embeddings', 512)
    if 'dropout_rate' not in configs:
        configs['dropout_rate'] = configs.get('hidden_dropout_prob')
    if 'segment_vocab_size' not in configs:
        configs['segment_vocab_size'] = configs.get('type_vocab_size', 2)

    models = {
        'bert': BERT
    }

    if is_string(model):
        model = model.lower()
        MODEL = models[model]
        if model.startswith('t5.1.1'):
            configs['version'] = 't5.1.1'
    else:
        MODEL = model

    application = application.lower()
    if application in ['lm', 'unilm'] and model in ['electra', 't5']:
        raise ValueError(
            '"%s" model can not be used as "%s" application.\n' %
            (model, application)
        )

    if application == 'lm':
        MODEL = extend_with_language_model(MODEL)
    elif application == 'unilm':
        MODEL = extend_with_custom_unified_language_model(MODEL)

    transformer = MODEL(**configs)
    transformer.build(**configs)

    if checkpoint_path is not None:
        transformer.load_weights_from_checkpoint(checkpoint_path)

    if return_keras_model:
        return transformer.model
    else:
        return transformer

model = build_transformer_model(
    config_path,
    checkpoint_path,
    application='unilm',
    keep_tokens=keep_tokens,
)

output = CrossEntropy(3)(model.inputs + model.outputs)

# Generate full-name
class FullNameGenerate(AutoRegressiveDecoder):
    @AutoRegressiveDecoder.wraps(default_rtype='probas')
    def predict(self, inputs, output_ids, states):
        token_ids, segment_ids, mask, tem_custom_ids = inputs
        token_ids = np.concatenate([token_ids, output_ids], 1)
        segment_ids = np.concatenate([segment_ids, np.ones_like(output_ids)], 1)
        tem_custom_ids = np.concatenate([tem_custom_ids, np.ones_like(output_ids)*3], 1)
        mask = generate_RanM(tem_custom_ids,mask_per)
        return self.last_token(model).predict([token_ids, segment_ids,mask])

    def generate(self, data_arr, lang, topk=1):
        max_c_len = maxlen - self.maxlen
        res=[]
        for arr in data_arr:
            text=data2sentences_DG(arr,lang)
            token_ids, segment_ids = M_encode_pred(text,max_c_len)
            custom_segment=get_custom_segment_pred(segment_ids)
            tem_custom_ids=custom_segment
            segment_ids=list(np.zeros_like(segment_ids))
            mask = generate_RanM([custom_segment],0)[0]
            output_ids = self.beam_search([token_ids, segment_ids, mask, tem_custom_ids],
                                          topk=topk)
            res.append(tokenizer.decode(output_ids))
        return res


fng = FullNameGenerate(start_id=None, end_id=tokenizer._token_end_id, maxlen=full_name_maxlen)

# show some examples
def just_show():
    for arr in data_list[test_index][:5]:
        text=data2sentences_DG(arr,'eng')
        full_name_pred=fng.generate(text)
        print(u'生成组织名:', full_name_pred)
        print(arr)
        print('')
    print()

def writeTestSetPredicion(epoch):
    res=fng.generate(data_list[test_index],'eng')
    tem_data=data_list[test_index]
    with open(firDir+'/TestPredictionOnEpoch_'+str(epoch)+'.txt','w',encoding='utf-8') as fw:
        for ind in range(len(tem_data)):
            fw.write('true: '+tem_data[ind][1]+'\t'+'predict: '+res[ind]+'\n')
    print('finish writing prediction of test set on epoch '+str(epoch))

def writeValSetPredicion(epoch):
    res=fng.generate(data_list[val_index],'eng')
    tem_data=data_list[val_index]
    with open(firDir+'/ValPredictionOnEpoch_'+str(epoch)+'.txt','w',encoding='utf-8') as fw:
        for ind in range(len(tem_data)):
            fw.write('true: '+tem_data[ind][1]+'\t'+'predict: '+res[ind]+'\n')
    print('finish writing prediction of validation set on epoch '+str(epoch))
    model.save_weights(firDir+'/LastModel.weights')

class Evaluator(keras.callbacks.Callback):
    def __init__(self):
        self.lowest = 1e10

    def on_epoch_end(self, epoch, logs=None):
        writeValSetPredicion(epoch+1)
        writeTestSetPredicion(epoch+1)

# divide dataset
kf = KFold(n_splits=5,shuffle=True,random_state=0)
train_index=0
val_test_index=0
val_index=0
test_index=0
num=0
for tr_index, vt_index in kf.split(data_list):
    num+=1
    train_index=tr_index
    val_test_index=vt_index
    # validation set and test set
    kff = KFold(n_splits=2,shuffle=True,random_state=1)
    for v_index, t_index in kff.split(val_test_index):
        val_index=val_test_index[v_index]
        test_index=val_test_index[t_index]
    evaluator = Evaluator()
    train_generator = data_generator(data_list[train_index], batch_size)
    model = Model(model.inputs, output)
    model.compile(optimizer=Adam(9e-6),metrics=['accuracy'])
    his=model.fit(
        train_generator.forfit(),
        steps_per_epoch=steps_per_epoch,
        epochs=epochs,
        callbacks=[evaluator]
    )
    break

val, val_ind=result_check_Val(firDir,epochs,'eng')
test=result_check_Test(firDir,'eng',val_ind)
metrics=['Levenshtein Distance','Normalized Levenshtein','Ratcliff-Obershelp Similarity','BLEU','Precision','Recall','F1-score']
for ind in range(len(test)):
    print(metrics[ind],test[ind])